==> v10.5.3.zip: brewman/brewman/brewman/routers/reports/reconcile.py <==
import uuid

from datetime import date, datetime
from typing import List, Optional

import brewman.schemas.reconcile as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session, joinedload
from sqlalchemy.sql.expression import and_, func, or_, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Reconcile)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    return schemas.Reconcile(
        startDate=get_start_date(request.session),
        finishDate=get_finish_date(request.session),
        account=None,
        body=[],
    )


@router.get("/{id_}", response_model=schemas.Reconcile)
def show_data(
    id_: uuid.UUID,
    request: Request,
    s: Optional[str] = None,
    f: Optional[str] = None,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    with SessionFuture() as db:
        account = db.execute(select(AccountBase).where(AccountBase.id == id_)).scalar_one()
        start_date = datetime.strptime(s or get_start_date(request.session), "%d-%b-%Y")
        finish_date = datetime.strptime(f or get_finish_date(request.session), "%d-%b-%Y")
        body = build_report(account.id, start_date, finish_date, db)
        set_period(start_date, finish_date, request.session)
        return schemas.Reconcile(
            startDate=start_date,
            finishDate=finish_date,
            account=schemas.AccountLink(id=account.id, name=account.name),
            body=body,
        )


def build_report(
    account_id: uuid.UUID, start_date: date, finish_date: date, db: Session
) -> List[schemas.ReconcileItem]:
    opening = opening_balance(account_id, start_date, db)
    body = [opening]

    query = db.execute(
        select(Voucher)
        .options(joinedload(Voucher.journals, innerjoin=True).joinedload(Journal.account, innerjoin=True))
        .where(
            Voucher.journals.any(Journal.account_id == account_id),
            or_(
                Voucher.is_reconciled == False,  # noqa: E712
                and_(
                    Voucher.reconcile_date >= start_date,
                    Voucher.reconcile_date <= finish_date,
                ),
            ),
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
        )
        .order_by(Voucher.is_reconciled, Voucher.reconcile_date, Voucher.last_edit_date)
    ).all()

    for voucher in query:
        debit = 0
        credit = 0
        journal_debit = 0
        name = ""
        for journal in voucher.journals:
            if journal.account_id == account_id:
                journal_debit = journal.debit
                if journal.debit == 1:
                    debit = journal.amount
                    credit = 0
                else:
                    credit = journal.amount
                    debit = 0
        for journal in voucher.journals:
            if journal.debit != journal_debit:
                name += "{0} / ".format(journal.account.name)
        name = name[:-3]
        body.append(
            schemas.ReconcileItem(
                id=voucher.id,
                date=voucher.date,
                name=name,
                type=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                debit=debit,
                credit=credit,
                isReconciled=voucher.is_reconciled,
                reconcileDate=voucher.reconcile_date,
            )
        )
    return body


def opening_balance(account_id: uuid.UUID, start_date: date, db: Session) -> schemas.ReconcileItem:
    opening = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .where(
            Voucher.reconcile_date < start_date,
            Voucher.is_reconciled == True,  # noqa: E712
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.account_id == account_id,
        )
    ).scalar()
    opening = 0 if opening is None else opening
    if opening < 0:
        credit = opening * -1
        debit = 0
    else:
        debit = opening
        credit = 0
    return schemas.ReconcileItem(
        date=start_date,
        id=None,
        name="Opening Balance",
        type="Opening Balance",
        narration="",
        debit=debit,
        credit=credit,
        running=opening,
        isReconciled=True,
        reconcileDate=start_date,
    )


@router.post("/{id_}", response_model=schemas.Reconcile)
def save(
    id_: uuid.UUID,
    data: schemas.Reconcile,
    request: Request,
    s: Optional[str] = None,
    f: Optional[str] = None,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    with SessionFuture() as db:
        account = db.execute(select(AccountBase).where(AccountBase.id == id_)).scalar_one()
        start_date = datetime.strptime(s or get_start_date(request.session), "%d-%b-%Y")
        finish_date = datetime.strptime(f or get_finish_date(request.session), "%d-%b-%Y")
        for item in data.body:
            if item.id_ is None:
                continue

            voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == item.id_)).scalar_one()
            voucher.is_reconciled = item.is_reconciled
            if item.reconcile_date is not None:
                voucher.reconcile_date = item.reconcile_date
            db.commit()
        body = build_report(account.id, start_date, finish_date, db)
        return schemas.Reconcile(
            startDate=start_date,
            finishDate=finish_date,
            account=schemas.AccountLink(id=account.id, name=account.name),
            body=body,
        )

==> v11.6.0.zip: brewman/brewman/brewman/routers/reports/reconcile.py <==
import uuid

from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session, joinedload
from sqlalchemy.sql.expression import and_, func, or_, select

import brewman.schemas.reconcile as schemas

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Reconcile)
def show_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    return schemas.Reconcile(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        account=None,
        body=[],
    )


@router.get("/{id_}", response_model=schemas.Reconcile)
def show_data(
    id_: uuid.UUID,
    request: Request,
    s: str | None = None,
    f: str | None = None,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    with SessionFuture() as db:
        account = db.execute(select(AccountBase).where(AccountBase.id == id_)).scalar_one()
        start_date = datetime.strptime(s or get_start_date(request.session), "%d-%b-%Y")
        finish_date = datetime.strptime(f or get_finish_date(request.session), "%d-%b-%Y")
        body = build_report(account.id, start_date, finish_date, db)
        set_period(start_date, finish_date, request.session)
        return schemas.Reconcile(
            start_date=start_date,
            finish_date=finish_date,
            account=schemas.AccountLink(id_=account.id, name=account.name),
            body=body,
        )


def build_report(
    account_id: uuid.UUID, start_date: date, finish_date: date, db: Session
) -> list[schemas.ReconcileItem]:
    opening = opening_balance(account_id, start_date, db)
    body = [opening]

    query = db.execute(
        select(Voucher)
        .options(joinedload(Voucher.journals, innerjoin=True).joinedload(Journal.account, innerjoin=True))
        .where(
            Voucher.journals.any(Journal.account_id == account_id),
            or_(
                Voucher.is_reconciled == False,  # noqa: E712
                and_(
                    Voucher.reconcile_date >= start_date,
                    Voucher.reconcile_date <= finish_date,
                ),
            ),
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
        )
        .order_by(Voucher.is_reconciled, Voucher.reconcile_date, Voucher.last_edit_date)
    ).all()

    for voucher in query:
        debit = Decimal(0)
        credit = Decimal(0)
        journal_debit = Decimal(0)
        name = ""
        for journal in voucher.journals:
            if journal.account_id == account_id:
                journal_debit = journal.debit
                if journal.debit == 1:
                    debit = journal.amount
                    credit = Decimal(0)
                else:
                    credit = journal.amount
                    debit = Decimal(0)
        for journal in voucher.journals:
            if journal.debit != journal_debit:
                name += f"{journal.account.name} / "
        name = name[:-3]
        body.append(
            schemas.ReconcileItem(
                id_=voucher.id,
                date_=voucher.date,
                name=name,
                type_=voucher.voucher_type.name.replace("_", " ").title(),
                narration=voucher.narration,
                url=[],
                posted=False,
                debit=debit,
                credit=credit,
                is_reconciled=voucher.is_reconciled,
                reconcile_date=voucher.reconcile_date,
            )
        )
    return body


def opening_balance(account_id: uuid.UUID, start_date: date, db: Session) -> schemas.ReconcileItem:
    opening = db.execute(
        select(func.sum(Journal.amount * Journal.debit))
        .join(Journal.voucher)
        .where(
            Voucher.reconcile_date < start_date,
            Voucher.is_reconciled == True,  # noqa: E712
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.account_id == account_id,
        )
    ).scalar()
    opening = Decimal(0) if opening is None else opening
    if opening < 0:
        credit = opening * -1
        debit = Decimal(0)
    else:
        debit = opening
        credit = Decimal(0)
    return schemas.ReconcileItem(
        date_=start_date,
        id_=None,
        name="Opening Balance",
        type_="Opening Balance",
        narration="",
        debit=debit,
        credit=credit,
        # running=opening,
        url=[],
        posted=False,
        is_reconciled=True,
        reconcile_date=start_date,
    )


@router.post("/{id_}", response_model=schemas.Reconcile)
def save(
    id_: uuid.UUID,
    data: schemas.Reconcile,
    request: Request,
    s: str | None = None,
    f: str | None = None,
    user: UserToken = Security(get_user, scopes=["reconcile"]),
) -> schemas.Reconcile:
    with SessionFuture() as db:
        account = db.execute(select(AccountBase).where(AccountBase.id == id_)).scalar_one()
        start_date = datetime.strptime(s or get_start_date(request.session), "%d-%b-%Y")
        finish_date = datetime.strptime(f or get_finish_date(request.session), "%d-%b-%Y")
        for item in data.body:
            if item.id_ is None:
                continue

            voucher: Voucher = db.execute(select(Voucher).where(Voucher.id == item.id_)).scalar_one()
            voucher.is_reconciled = item.is_reconciled
            if item.reconcile_date is not None:
                voucher.reconcile_date = item.reconcile_date
            db.commit()
        body = build_report(account.id, start_date, finish_date, db)
        return schemas.Reconcile(
            start_date=start_date,
            finish_date=finish_date,
            account=schemas.AccountLink(id_=account.id, name=account.name),
            body=body,
        )
